/*
** Module   : SequenceManager.java
** Abstract : Static abstract class that handles sequence functions and 
**            statements from Progress / OpenEdge 10.x. The actual 
**            implementation is delegated to one of the derived classes that
**            implements a strategy of handling the persistence of the 
**            sequences.
**
** Copyright (c) 2012, Golden Code Development Corporation.
** ALL RIGHTS RESERVED. Use is subject to license terms.
**
**           Golden Code Development Corporation
**                      CONFIDENTIAL
**
** -#- -I- --Date-- ---------------------Description---------------------
** 001 OM  20121116 Created initial version.
*/

package com.goldencode.p2j.persist;

import java.util.*;
import org.hibernate.dialect.Dialect;
import com.goldencode.p2j.persist.dialect.*;
import com.goldencode.p2j.persist.sequence.*;
import com.goldencode.p2j.util.*;

/**
 * A class that abstracts all sequence handling functions and statements,
 * including static dynamic forms.
 * 
 * At the moment of initial commit we believe (and hope) that runtime
 * implementations of the statement/functions should leverage dialect
 * infrastructure in Hibernate if possible (if we find it is feasible to
 * use native sequences).
 * 
 * // db support for sequences should be tested with Dialect.supportsSequences()
 * 
 * The following schema properties (with the noted settings) must be 
 * supported by schema conversion:
 * <ul>
 *    <li>initial value of 0, 1 or 100
 *    <li>increment of 1
 *    <li>cycle on limit of no and yes
 *    <li>minimum value of 0, 1 or 100
 *    <li>maximum value (it is only sometimes explicitly specified) of 9999, 
 *          999999999 or 2000000000
 * </ul>
 */
public abstract class SequenceManager
{
   /**
    * The list of all known sequences in a double hashed Map.
    * <p>
    * <i>I wonder if the first level uses the sequence names and the second 
    * the database. This makes it easier to look for a sequence when the name 
    * of the database is default/unknown. </i>
    */
   private static HashMap<String, HashMap<String, Sequence>> sequences =
      new HashMap<String, HashMap<String, Sequence>>();

   /** 
    * Internal data structure with all SequenceHandler used
    * The keys are logical database names.
    */
   private static HashMap<String, SequenceHandler> seqHandlers =
      new HashMap<String, SequenceHandler>();

   /**
    * Maps the <code>CURRENT-VALUE(seqName)</code> function from Progress as
    * it is defined in Open Edge 10.x.
    * Returns the current value of a sequence defined in the Data
    * Dictionary.
    *
    * @param   seqName
    *          An identifier that specifies the name of a sequence defined
    *          in the Data Dictionary.
    *
    * @return
    *       The current value of a sequence can be one of the following:
    *       <ul>
    *          <li>The initial value specified in the Data Dictionary.
    *          <li>The last value set with either the
    *             <code>CURRENT-VALUE</code> statement or the 
    *             <code>NEXT-VALUE</code> function.
    *          <li>The unknown value (<code>?</code>) if the sequence has 
    *             exceeded its minimum or maximum and is not cycling.
    *       </ul>
    */
   public static int64 currentValue(String seqName)
   {
      String ldbName = getDatabaseForSequence(seqName);
      if (ldbName == null)
      {
         // database could not be identified
         return new int64();
      }
      return getSequenceHandler(ldbName).getCurrentValue(seqName, ldbName);
   }

   /**
    * Maps the <code>CURRENT-VALUE(seqName, ldbName)</code> function from 
    * Progress as it is defined in Open Edge 10.x.
    * Returns the current value of the sequence defined in the Data
    * Dictionary.
    *
    * @param   seqName
    *          An identifier that specifies the name of a sequence defined
    *          in the Data Dictionary.
    * @param ldbName
    *          An identifier that specifies the logical name of the database
    *          in which the sequence is defined. The database must be
    *          connected. You can omit this parameter if the sequence name is
    *          unambiguous. If a sequence with this name exists in more than
    *          one connected database, then you must specify
    *          <code>ldbName</code>.
    *
    * @return
    *       The current value of a sequence can be one of the following:
    *       <ul>
    *          <li>The initial value specified in the Data Dictionary.
    *          <li>The last value set with either the
    *             <code>CURRENT-VALUE</code> statement or the 
    *             <code>NEXT-VALUE</code> function.
    *          <li>The unknown value (<code>?</code>) if the sequence has 
    *             exceeded its minimum or maximum and is not cycling.
    *       </ul>
    */
   public static int64 currentValue(String seqName, String ldbName)
   {
      return getSequenceHandler(ldbName).getCurrentValue(seqName, ldbName);
   }

   /**
    * Maps the <code>NEXT-VALUE(seqName)</code> function from Progress as it
    * is defined in Open Edge 10.x.
    * Returns the next value of the sequence, incremented by the
    * positive or negative value defined in the Data Dictionary.
    *
    * @param   seqName
    *          An identifier that specifies the name of a sequence defined
    *          in the Data Dictionary.
    *
    * @return
    *       The next value of a sequence can be one of the following:
    *       <ul>
    *          <li>The value obtained by incremented the current value of the 
    *             sequence with its increment if the value does not cross the
    *             upper/lower limits. In the case of cycling sequences the
    *             sequence value is warp around. 
    *          <li>The unknown value (<code>?</code>) if the sequence has 
    *             exceeded its minimum or maximum and is not cycling.
    *       </ul>
    */
   public static int64 nextValue(String seqName)
   {
      String ldbName = getDatabaseForSequence(seqName);
      if (ldbName == null)
      {
         // database could not be identified
         return new int64();
      }
      return getSequenceHandler(ldbName).getNextValue(seqName, ldbName);
   }

   /**
    * Maps the <code>NEXT-VALUE(seqName, ldbName)</code> function from
    * Progress as it is defined in Open Edge 10.x.
    * Returns the next value of the sequence, incremented by the
    * positive or negative value defined in the Data Dictionary.
    *
    * @param   seqName
    *          An identifier that specifies the name of a sequence defined
    *          in the Data Dictionary.
    * @param ldbName
    *          An identifier that specifies the logical name of the database
    *          in which the sequence is defined. The database must be
    *          connected. If multiple databases are connected, you can omit
    *          this parameter if you specify a sequence that is unique to
    *          one of the databases.
    *
    * @return
    *       The next value of a sequence can be one of the following:
    *       <ul>
    *          <li>The initial value specified in the Data Dictionary.
    *          <li>The last value set with either the
    *             <code>CURRENT-VALUE</code> statement or the 
    *             <code>NEXT-VALUE</code> function.
    *          <li>The unknown value (<code>?</code>) if the sequence has 
    *             exceeded its minimum or maximum and is not cycling.
    *       </ul>
    */
   public static int64 nextValue(String seqName, String ldbName)
   {
      return getSequenceHandler(ldbName).getNextValue(seqName, ldbName);
   }

   /**
    * Maps the <code>DYNAMIC-CURRENT-VALUE(seqName, ldbName)</code> function
    * from Progress as it is defined in Open Edge 10.x.
    * Returns the current value of the sequence defined in the Data
    * Dictionary.
    *
    * @param   seqName
    *          The name of a sequence defined in the Data Dictionary.
    * @param   ldbName
    *          The logical name of the database in which the sequence is 
    *          defined. The database must be connected.
    *
    * @return
    *       The current value of a sequence can be one of the following:
    *       <ul>
    *          <li>The initial value specified in the Data Dictionary.
    *          <li>The last value set with either the
    *             <code>CURRENT-VALUE</code> statement or the 
    *             <code>NEXT-VALUE</code> function.
    *          <li>The unknown value (<code>?</code>) if the sequence has 
    *             exceeded its minimum or maximum and is not cycling.
    *       </ul>
    */
   public static int64 dynamicCurrentValue(String seqName, String ldbName)
   {
      return getSequenceHandler(ldbName).getCurrentValue(seqName, ldbName);
   }

   /**
    * Maps the <code>DYNAMIC-CURRENT-VALUE(seqName, ldbName)</code> function
    * from Progress as it is defined in Open Edge 10.x.
    * Returns the current value of the sequence defined in the Data
    * Dictionary.
    *
    * @param   seqName
    *          The name of a sequence defined in the Data Dictionary.
    * @param   ldbName
    *          A character expression that evaluates to a logical name of 
    *          the database in which the sequence is defined. The database 
    *          must be connected.
    *
    * @return
    *       The current value of a sequence can be one of the following:
    *       <ul>
    *          <li>The initial value specified in the Data Dictionary.
    *          <li>The last value set with either the
    *             <code>CURRENT-VALUE</code> statement or the 
    *             <code>NEXT-VALUE</code> function.
    *          <li>The unknown value (<code>?</code>) if the sequence has 
    *             exceeded its minimum or maximum and is not cycling.
    *       </ul>
    */
   public static int64 dynamicCurrentValue(String seqName,
                                           character ldbName)
   {
      String _ldbName = ldbName.toJavaType();
      return getSequenceHandler(_ldbName).getCurrentValue(seqName, _ldbName);
   }

   /**
    * Maps the <code>DYNAMIC-CURRENT-VALUE(seqName, ldbName)</code> function
    * from Progress as it is defined in Open Edge 10.x.
    * Returns the current value of the sequence defined in the Data
    * Dictionary.
    *
    * @param   seqName
    *          A character expression that evaluates to the name of a
    *          sequence defined in the Data Dictionary.
    * @param   ldbName
    *          The logical name of the database in which the sequence is 
    *          defined. The database must be connected.
    *
    * @return
    *       The current value of a sequence can be one of the following:
    *       <ul>
    *          <li>The initial value specified in the Data Dictionary.
    *          <li>The last value set with either the
    *             <code>CURRENT-VALUE</code> statement or the 
    *             <code>NEXT-VALUE</code> function.
    *          <li>The unknown value (<code>?</code>) if the sequence has 
    *             exceeded its minimum or maximum and is not cycling.
    *       </ul>
    */
   public static int64 dynamicCurrentValue(character seqName, 
                                           String ldbName)
   {
      return getSequenceHandler(ldbName)
            .getCurrentValue(seqName.toJavaType(), ldbName);
   }

   /**
    * Maps the <code>DYNAMIC-CURRENT-VALUE(seqName, ldbName)</code> function
    * from Progress as it is defined in Open Edge 10.x.
    * Returns the current value of the sequence defined in the Data
    * Dictionary.
    *
    * @param   seqName
    *          A character expression that evaluates to the name of a
    *          sequence defined in the Data Dictionary.
    * @param ldbName
    *          A character expression that evaluates to a logical name of
    *          the database in which the sequence is defined. The database
    *          must be connected.
    *
    * @return
    *       The current value of a sequence can be one of the following:
    *       <ul>
    *          <li>The initial value specified in the Data Dictionary.
    *          <li>The last value set with either the
    *             <code>CURRENT-VALUE</code> statement or the 
    *             <code>NEXT-VALUE</code> function.
    *          <li>The unknown value (<code>?</code>) if the sequence has 
    *             exceeded its minimum or maximum and is not cycling.
    *       </ul>
    */
   public static int64 dynamicCurrentValue(character seqName,
                                           character ldbName)
   {
      String _ldbName = ldbName.toJavaType();
      return getSequenceHandler(_ldbName)
            .getCurrentValue(seqName.toJavaType(), _ldbName);
   }

   /**
    * Maps the <code>DYNAMIC-NEXT-VALUE(seqName, ldbName)</code> function
    * from Progress as it is defined in Open Edge 10.x.
    * Returns the next value of the sequence, incremented by the
    * positive or negative value defined in the Data Dictionary.
    *
    * @param   seqName
    *          The name of a sequence defined in the Data Dictionary.
    * @param   ldbName
    *          The logical name of the database in which the sequence is 
    *          defined. The database must be connected.
    *
    * @return
    *       The next value of a sequence can be one of the following:
    *       <ul>
    *          <li>The value obtained by incremented the current value of the 
    *             sequence with its increment if the value does not cross the
    *             upper/lower limits. In the case of cycling sequences the
    *             sequence value is warp around. 
    *          <li>The unknown value (<code>?</code>) if the sequence has 
    *             exceeded its minimum or maximum and is not cycling.
    *       </ul>
    */
   public static int64 dynamicNextValue(String seqName,
                                        String ldbName)
   {
      return getSequenceHandler(ldbName).getNextValue(seqName, ldbName);
   }

   /**
    * Maps the <code>DYNAMIC-NEXT-VALUE(seqName, ldbName)</code> function
    * from Progress as it is defined in Open Edge 10.x.
    * Returns the next value of the sequence, incremented by the
    * positive or negative value defined in the Data Dictionary.
    *
    * @param   seqName
    *          The name of a sequence defined in the Data Dictionary.
    * @param   ldbName
    *          A character expression that evaluates to a logical name of
    *          the database in which the sequence is defined. The database
    *          must be connected.
    *
    * @return
    *       The next value of a sequence can be one of the following:
    *       <ul>
    *          <li>The value obtained by incremented the current value of the 
    *             sequence with its increment if the value does not cross the
    *             upper/lower limits. In the case of cycling sequences the
    *             sequence value is warp around. 
    *          <li>The unknown value (<code>?</code>) if the sequence has 
    *             exceeded its minimum or maximum and is not cycling.
    *       </ul>
    */
   public static int64 dynamicNextValue(String seqName,
                                        character ldbName)
   {
      String _ldbName = ldbName.toJavaType();
      return getSequenceHandler(_ldbName)
            .getNextValue(seqName, _ldbName);
   }

   /**
    * Maps the <code>DYNAMIC-NEXT-VALUE(seqName, ldbName)</code> function
    * from Progress as it is defined in Open Edge 10.x.
    * Returns the next value of the sequence, incremented by the
    * positive or negative value defined in the Data Dictionary.
    *
    * @param   seqName
    *          A character expression that evaluates to the name of a
    *          sequence defined in the Data Dictionary.
    * @param   ldbName
    *          The logical name of the database in which the sequence is 
    *          defined. The database must be connected.
    *
    * @return
    *       The next value of a sequence can be one of the following:
    *       <ul>
    *          <li>The value obtained by incremented the current value of the 
    *             sequence with its increment if the value does not cross the
    *             upper/lower limits. In the case of cycling sequences the
    *             sequence value is warp around. 
    *          <li>The unknown value (<code>?</code>) if the sequence has 
    *             exceeded its minimum or maximum and is not cycling.
    *       </ul>
    */
   public static int64 dynamicNextValue(character seqName,
                                        String ldbName)
   {
      return getSequenceHandler(ldbName)
            .getNextValue(seqName.toJavaType(), ldbName);
   }

   /**
    * Maps the <code>DYNAMIC-NEXT-VALUE(seqName, ldbName)</code> function
    * from Progress as it is defined in Open Edge 10.x.
    * Returns the next value of the sequence, incremented by the
    * positive or negative value defined in the Data Dictionary.
    *
    * @param   seqName
    *          A character expression that evaluates to the name of a
    *          sequence defined in the Data Dictionary.
    * @param ldbName
    *          A character expression that evaluates to a logical name of
    *          the database in which the sequence is defined. The database
    *          must be connected.
    *
    * @return
    *       The next value of a sequence can be one of the following:
    *       <ul>
    *          <li>The value obtained by incremented the current value of the 
    *             sequence with its increment if the value does not cross the
    *             upper/lower limits. In the case of cycling sequences the
    *             sequence value is warp around. 
    *          <li>The unknown value (<code>?</code>) if the sequence has 
    *             exceeded its minimum or maximum and is not cycling.
    *       </ul>
    */
   public static int64 dynamicNextValue(character seqName,
                                        character ldbName)
   {
      String _ldbName = ldbName.toJavaType();
      return getSequenceHandler(_ldbName)
            .getNextValue(seqName.toJavaType(), _ldbName);
   }

   /**
    * Maps the <code>CURRENT-VALUE(seqName) = expr</code> statement from
    * Progress as it is defined in Open Edge 10.x.
    * Resets the current value of the sequence defined in the Data
    * Dictionary.
    *
    * @param   seqName
    *          An identifier that specifies the name of a sequence defined
    *          in the Data Dictionary.
    * @param   newValue
    *          An integer expression assigned as the current value of the
    *          specified sequence. If expression is outside the boundary set
    *          by the initial value (at one end) and the lower limit or upper
    *          limit (at the other end) for the sequence, Progress returns an
    *          error, and the sequence value remains unchanged.
    */
   public static void setValue(String seqName, integer newValue)
   {
      String ldbName = getDatabaseForSequence(seqName);
      if (ldbName == null)
      {
         // database could not be identified
         return;
      }
      getSequenceHandler(ldbName)
         .setValue(seqName, ldbName, newValue.longValue());
   }

   /**
    * Maps the <code>CURRENT-VALUE(seqName) = expr</code> statement from
    * Progress as it is defined in Open Edge 10.x.
    * Resets the current value of the sequence defined in the Data
    * Dictionary.
    *
    * @param   seqName
    *          An identifier that specifies the name of a sequence defined
    *          in the Data Dictionary.
    * @param   newValue
    *          An integer expression assigned as the current value of the
    *          specified sequence. If expression is outside the boundary set
    *          by the initial value (at one end) and the lower limit or upper
    *          limit (at the other end) for the sequence, Progress returns an
    *          error, and the sequence value remains unchanged.
    */
   public static void setValue(String seqName, int newValue)
   {
      String ldbName = getDatabaseForSequence(seqName);
      if (ldbName == null)
      {
         // database could not be identified
         return;
      }
      getSequenceHandler(ldbName).setValue(seqName, ldbName, newValue);
   }

   /**
    * Maps the <code>CURRENT-VALUE(seqName, ldbName) = expr</code> statement
    * from Progress as it is defined in Open Edge 10.x.
    * Resets the current value of the sequence defined in the Data
    * Dictionary.
    *
    * @param   seqName
    *          An identifier that specifies the name of a sequence defined
    *          in the Data Dictionary.
    * @param   ldbName
    *          An identifier that specifies the logical name of the database
    *          in which the sequence is defined. The database must be
    *          connected. You can omit this parameter if the sequence name is
    *          unambiguous. If more than one connected database has a sequence
    *          with given name, then you must supply <code>ldbName</code>.
    * @param   newValue
    *          An integer expression assigned as the current value of the
    *          specified sequence. If expression is outside the boundary set
    *          by the initial value (at one end) and the lower limit or upper
    *          limit (at the other end) for the sequence, Progress returns an
    *          error, and the sequence value remains unchanged.
    */
   public static void setValue(String seqName,
                               int newValue,
                               String ldbName)
   {
      getSequenceHandler(ldbName).setValue(seqName, ldbName, newValue);
   }

   /**
    * Maps the <code>CURRENT-VALUE(seqName, ldbName) = expr</code> statement
    * from Progress as it is defined in Open Edge 10.x.
    * Resets the current value of the sequence defined in the Data
    * Dictionary.
    *
    * @param   seqName
    *          An identifier that specifies the name of a sequence defined
    *          in the Data Dictionary.
    * @param   ldbName
    *          An identifier that specifies the logical name of the database
    *          in which the sequence is defined. The database must be
    *          connected. You can omit this parameter if the sequence name is
    *          unambiguous. If more than one connected database has a sequence
    *          with given name, then you must supply <code>ldbName</code>.
    * @param   newValue
    *          An integer expression assigned as the current value of the
    *          specified sequence. If expression is outside the boundary set
    *          by the initial value (at one end) and the lower limit or upper
    *          limit (at the other end) for the sequence, Progress returns an
    *          error, and the sequence value remains unchanged.
    */
   public static void setValue(String seqName,
                               integer newValue,
                               String ldbName)
   {
      if (newValue == null || newValue.isUnknown())
      {
         // TODO: check this behavior
         throw new RuntimeException(
            "Unable to evaluate field for assignment. (143)");
      }
      getSequenceHandler(ldbName)
         .setValue(seqName, ldbName, newValue.longValue());
   }

   /**
    * Maps the <code>CURRENT-VALUE(seqName) = expr</code> statement from
    * Progress as it is defined in Open Edge 10.x.
    * Resets the current value of the sequence defined in the Data
    * Dictionary.
    *
    * @param   seqName
    *          An identifier that specifies the name of a sequence defined
    *          in the Data Dictionary.
    * @param   newValue
    *          An integer expression assigned as the current value of the
    *          specified sequence. If expression is outside the boundary set
    *          by the initial value (at one end) and the lower limit or upper
    *          limit (at the other end) for the sequence, Progress returns an
    *          error, and the sequence value remains unchanged.
    */
   public static void setValue(String seqName, int64 newValue)
   {
      String ldbName = getDatabaseForSequence(seqName);
      if (ldbName == null)
      {
         // database could not be identified
         return;
      }
      getSequenceHandler(ldbName)
         .setValue(seqName, ldbName, newValue.longValue());
   }

   /**
    * Maps the <code>CURRENT-VALUE(seqName) = expr</code> statement from
    * Progress as it is defined in Open Edge 10.x.
    * Resets the current value of the sequence defined in the Data
    * Dictionary.
    *
    * @param   seqName
    *          An identifier that specifies the name of a sequence defined
    *          in the Data Dictionary.
    * @param   newValue
    *          An integer expression assigned as the current value of the
    *          specified sequence. If expression is outside the boundary set
    *          by the initial value (at one end) and the lower limit or upper
    *          limit (at the other end) for the sequence, Progress returns an
    *          error, and the sequence value remains unchanged.
    */
   public static void setValue(String seqName, long newValue)
   {
      String ldbName = getDatabaseForSequence(seqName);
      if (ldbName == null)
      {
         // database could not be identified
         return;
      }
      getSequenceHandler(ldbName).setValue(seqName, ldbName, newValue);
   }

   /**
    * Maps the <code>CURRENT-VALUE(seqName, ldbName) = expr</code> statement
    * from Progress as it is defined in Open Edge 10.x.
    * Resets the current value of the sequence defined in the Data
    * Dictionary.
    *
    * @param   seqName
    *          An identifier that specifies the name of a sequence defined
    *          in the Data Dictionary.
    * @param   ldbName
    *          An identifier that specifies the logical name of the database
    *          in which the sequence is defined. The database must be
    *          connected. You can omit this parameter if the sequence name is
    *          unambiguous. If more than one connected database has a sequence
    *          with given name, then you must supply <code>ldbName</code>.
    * @param   newValue
    *          An integer expression assigned as the current value of the
    *          specified sequence. If expression is outside the boundary set
    *          by the initial value (at one end) and the lower limit or upper
    *          limit (at the other end) for the sequence, Progress returns an
    *          error, and the sequence value remains unchanged.
    */
   public static void setValue(String seqName,
                               long newValue,
                               String ldbName)
   {
      getSequenceHandler(ldbName).setValue(seqName, ldbName, newValue);
   }

   /**
    * Maps the <code>CURRENT-VALUE(seqName, ldbName) = expr</code> statement
    * from Progress as it is defined in Open Edge 10.x.
    * Resets the current value of the sequence defined in the Data
    * Dictionary.
    *
    * @param   seqName
    *          An identifier that specifies the name of a sequence defined
    *          in the Data Dictionary.
    * @param   ldbName
    *          An identifier that specifies the logical name of the database
    *          in which the sequence is defined. The database must be
    *          connected. You can omit this parameter if the sequence name is
    *          unambiguous. If more than one connected database has a sequence
    *          with given name, then you must supply <code>ldbName</code>.
    * @param   newValue
    *          An integer expression assigned as the current value of the
    *          specified sequence. If expression is outside the boundary set
    *          by the initial value (at one end) and the lower limit or upper
    *          limit (at the other end) for the sequence, Progress returns an
    *          error, and the sequence value remains unchanged.
    */
   public static void setValue(String seqName,
                               int64 newValue,
                               String ldbName)
   {
      getSequenceHandler(ldbName)
         .setValue(seqName, ldbName, newValue.longValue());
   }

   /**
    * Maps the <code>DYNAMIC-CURRENT-VALUE(seqName, ldbName) = expr</code> 
    * statement from Progress as it is defined in Open Edge 10.x.
    * Resets the current value of the sequence defined in the Data
    * Dictionary.
    *
    * @param   seqName
    *          An identifier that specifies the name of a sequence defined
    *          in the Data Dictionary.
    * @param   ldbName
    *          An identifier that specifies the logical name of the database
    *          in which the sequence is defined. The database must be
    *          connected. You can omit this parameter if the sequence name is
    *          unambiguous. If more than one connected database has a sequence
    *          with given name, then you must supply <code>ldbName</code>.
    * @param   newValue
    *          An integer expression assigned as the current value of the
    *          specified sequence. If expression is outside the boundary set
    *          by the initial value (at one end) and the lower limit or upper
    *          limit (at the other end) for the sequence, Progress returns an
    *          error, and the sequence value remains unchanged.
    */
   public static void dynamicSetValue(String seqName,
                                      int newValue,
                                      String ldbName)
   {
      getSequenceHandler(ldbName).setValue(seqName, ldbName, newValue);
   }

   /**
    * Maps the <code>DYNAMIC-CURRENT-VALUE(seqName, ldbName) = expr</code> 
    * statement from Progress as it is defined in Open Edge 10.x.
    * Resets the current value of the sequence defined in the Data
    * Dictionary.
    *
    * @param   seqName
    *          An identifier that specifies the name of a sequence defined
    *          in the Data Dictionary.
    * @param   ldbName
    *          An identifier that specifies the logical name of the database
    *          in which the sequence is defined. The database must be
    *          connected. You can omit this parameter if the sequence name is
    *          unambiguous. If more than one connected database has a sequence
    *          with given name, then you must supply <code>ldbName</code>.
    * @param   newValue
    *          An integer expression assigned as the current value of the
    *          specified sequence. If expression is outside the boundary set
    *          by the initial value (at one end) and the lower limit or upper
    *          limit (at the other end) for the sequence, Progress returns an
    *          error, and the sequence value remains unchanged.
    */
   public static void dynamicSetValue(String seqName,
                                      int newValue,
                                      character ldbName)
   {
      String _ldbName = ldbName.toJavaType();
      getSequenceHandler(_ldbName).setValue(seqName, _ldbName, newValue);
   }

   /**
    * Maps the <code>DYNAMIC-CURRENT-VALUE(seqName, ldbName) = expr</code>
    * statement from Progress as it is defined in Open Edge 10.x.
    * Resets the current value of the sequence defined in the Data
    * Dictionary.
    *
    * @param   seqName
    *          A character expression that evaluates to the name of a 
    *          sequence defined in the Data Dictionary.
    * @param   ldbName
    *          A character expression that evaluates to the name of a database
    *          in which the sequence is defined. The database must be
    *          connected. You can omit this parameter if the sequence name is
    *          unambiguous. If more than one connected database has a sequence
    *          with given name, then you must supply <code>ldbName</code>.
    * @param   newValue
    *          An integer expression assigned as the current value of the
    *          specified sequence. If expression is outside the boundary set
    *          by the initial value (at one end) and the lower limit or upper
    *          limit (at the other end) for the sequence, Progress returns an
    *          error, and the sequence value remains unchanged.
    */
   public static void dynamicSetValue(String seqName,
                                      integer newValue,
                                      String ldbName)
   {
      getSequenceHandler(ldbName).setValue(seqName,
                                           ldbName,
                                           newValue.longValue());
   }

   /**
    * Maps the <code>DYNAMIC-CURRENT-VALUE(seqName, ldbName) = expr</code> 
    * statement from Progress as it is defined in Open Edge 10.x.
    * Resets the current value of the sequence defined in the Data
    * Dictionary.
    *
    * @param   seqName
    *          An identifier that specifies the name of a sequence defined
    *          in the Data Dictionary.
    * @param   ldbName
    *          An identifier that specifies the logical name of the database
    *          in which the sequence is defined. The database must be
    *          connected. You can omit this parameter if the sequence name is
    *          unambiguous. If more than one connected database has a sequence
    *          with given name, then you must supply <code>ldbName</code>.
    * @param   newValue
    *          An integer expression assigned as the current value of the
    *          specified sequence. If expression is outside the boundary set
    *          by the initial value (at one end) and the lower limit or upper
    *          limit (at the other end) for the sequence, Progress returns an
    *          error, and the sequence value remains unchanged.
    */
   public static void dynamicSetValue(String seqName,
                                      integer newValue,
                                      character ldbName)
   {
      String _ldbName = ldbName.toJavaType();
      getSequenceHandler(_ldbName).setValue(seqName,
                                            _ldbName,
                                            newValue.longValue());
   }

   /**
    * Maps the <code>DYNAMIC-CURRENT-VALUE(seqName, ldbName) = expr</code>
    * statement from Progress as it is defined in Open Edge 10.x.
    * Resets the current value of the sequence defined in the Data
    * Dictionary.
    *
    * @param   seqName
    *          A character expression that evaluates to the name of a 
    *          sequence defined in the Data Dictionary.
    * @param   ldbName
    *          A character expression that evaluates to the name of a database
    *          in which the sequence is defined. The database must be
    *          connected. You can omit this parameter if the sequence name is
    *          unambiguous. If more than one connected database has a sequence
    *          with given name, then you must supply <code>ldbName</code>.
    * @param   newValue
    *          An integer expression assigned as the current value of the
    *          specified sequence. If expression is outside the boundary set
    *          by the initial value (at one end) and the lower limit or upper
    *          limit (at the other end) for the sequence, Progress returns an
    *          error, and the sequence value remains unchanged.
    */
   public static void dynamicSetValue(character seqName,
                                      int newValue,
                                      String ldbName)
   {
      getSequenceHandler(ldbName).setValue(seqName.toJavaType(), 
                                           ldbName,
                                           newValue);
   }

   /**
    * Maps the <code>DYNAMIC-CURRENT-VALUE(seqName, ldbName) = expr</code> 
    * statement from Progress as it is defined in Open Edge 10.x.
    * Resets the current value of the sequence defined in the Data
    * Dictionary.
    *
    * @param   seqName
    *          An identifier that specifies the name of a sequence defined
    *          in the Data Dictionary.
    * @param   ldbName
    *          An identifier that specifies the logical name of the database
    *          in which the sequence is defined. The database must be
    *          connected. You can omit this parameter if the sequence name is
    *          unambiguous. If more than one connected database has a sequence
    *          with given name, then you must supply <code>ldbName</code>.
    * @param   newValue
    *          An integer expression assigned as the current value of the
    *          specified sequence. If expression is outside the boundary set
    *          by the initial value (at one end) and the lower limit or upper
    *          limit (at the other end) for the sequence, Progress returns an
    *          error, and the sequence value remains unchanged.
    */
   public static void dynamicSetValue(character seqName,
                                      int newValue,
                                      character ldbName)
   {
      String _ldbName = ldbName.toJavaType();
      getSequenceHandler(_ldbName)
         .setValue(seqName.toJavaType(), _ldbName, newValue);
   }

   /**
    * Maps the <code>DYNAMIC-CURRENT-VALUE(seqName, ldbName) = expr</code>
    * statement from Progress as it is defined in Open Edge 10.x.
    * Resets the current value of the sequence defined in the Data
    * Dictionary.
    *
    * @param   seqName
    *          A character expression that evaluates to the name of a 
    *          sequence defined in the Data Dictionary.
    * @param   ldbName
    *          A character expression that evaluates to the name of a database
    *          in which the sequence is defined. The database must be
    *          connected. You can omit this parameter if the sequence name is
    *          unambiguous. If more than one connected database has a sequence
    *          with given name, then you must supply <code>ldbName</code>.
    * @param   newValue
    *          An integer expression assigned as the current value of the
    *          specified sequence. If expression is outside the boundary set
    *          by the initial value (at one end) and the lower limit or upper
    *          limit (at the other end) for the sequence, Progress returns an
    *          error, and the sequence value remains unchanged.
    */
   public static void dynamicSetValue(character seqName,
                                      integer newValue,
                                      String ldbName)
   {
      getSequenceHandler(ldbName).setValue(seqName.toJavaType(),
                                           ldbName,
                                           newValue.longValue());
   }

   /**
    * Maps the <code>DYNAMIC-CURRENT-VALUE(seqName, ldbName) = expr</code>
    * statement from Progress as it is defined in Open Edge 10.x.
    * Resets the current value of the sequence defined in the Data
    * Dictionary.
    *
    * @param   seqName
    *          A character expression that evaluates to the name of a 
    *          sequence defined in the Data Dictionary.
    * @param   ldbName
    *          A character expression that evaluates to the name of a database
    *          in which the sequence is defined. The database must be
    *          connected. You can omit this parameter if the sequence name is
    *          unambiguous. If more than one connected database has a sequence
    *          with given name, then you must supply <code>ldbName</code>.
    * @param   newValue
    *          An integer expression assigned as the current value of the
    *          specified sequence. If expression is outside the boundary set
    *          by the initial value (at one end) and the lower limit or upper
    *          limit (at the other end) for the sequence, Progress returns an
    *          error, and the sequence value remains unchanged.
    */
   public static void dynamicSetValue(character seqName,
                                      integer newValue,
                                      character ldbName)
   {
      String _ldbName = ldbName.toJavaType();
      getSequenceHandler(_ldbName)
         .setValue(seqName.toJavaType(), _ldbName, newValue.longValue());
   }

   /**
    * Maps the <code>DYNAMIC-CURRENT-VALUE(seqName, ldbName) = expr</code> 
    * statement from Progress as it is defined in Open Edge 10.x.
    * Resets the current value of the sequence defined in the Data
    * Dictionary.
    *
    * @param   seqName
    *          An identifier that specifies the name of a sequence defined
    *          in the Data Dictionary.
    * @param   ldbName
    *          An identifier that specifies the logical name of the database
    *          in which the sequence is defined. The database must be
    *          connected. You can omit this parameter if the sequence name is
    *          unambiguous. If more than one connected database has a sequence
    *          with given name, then you must supply <code>ldbName</code>.
    * @param   newValue
    *          An integer expression assigned as the current value of the
    *          specified sequence. If expression is outside the boundary set
    *          by the initial value (at one end) and the lower limit or upper
    *          limit (at the other end) for the sequence, Progress returns an
    *          error, and the sequence value remains unchanged.
    */
   public static void dynamicSetValue(String seqName,
                                      long newValue,
                                      String ldbName)
   {
      getSequenceHandler(ldbName).setValue(seqName, ldbName, newValue);
   }

   /**
    * Maps the <code>DYNAMIC-CURRENT-VALUE(seqName, ldbName) = expr</code> 
    * statement from Progress as it is defined in Open Edge 10.x.
    * Resets the current value of the sequence defined in the Data
    * Dictionary.
    *
    * @param   seqName
    *          An identifier that specifies the name of a sequence defined
    *          in the Data Dictionary.
    * @param   ldbName
    *          An identifier that specifies the logical name of the database
    *          in which the sequence is defined. The database must be
    *          connected. You can omit this parameter if the sequence name is
    *          unambiguous. If more than one connected database has a sequence
    *          with given name, then you must supply <code>ldbName</code>.
    * @param   newValue
    *          An integer expression assigned as the current value of the
    *          specified sequence. If expression is outside the boundary set
    *          by the initial value (at one end) and the lower limit or upper
    *          limit (at the other end) for the sequence, Progress returns an
    *          error, and the sequence value remains unchanged.
    */
   public static void dynamicSetValue(String seqName,
                                      long newValue,
                                      character ldbName)
   {
      String _ldbName = ldbName.toJavaType();
      getSequenceHandler(_ldbName).setValue(seqName, _ldbName, newValue);
   }

   /**
    * Maps the <code>DYNAMIC-CURRENT-VALUE(seqName, ldbName) = expr</code>
    * statement from Progress as it is defined in Open Edge 10.x.
    * Resets the current value of the sequence defined in the Data
    * Dictionary.
    *
    * @param   seqName
    *          A character expression that evaluates to the name of a 
    *          sequence defined in the Data Dictionary.
    * @param   ldbName
    *          A character expression that evaluates to the name of a database
    *          in which the sequence is defined. The database must be
    *          connected. You can omit this parameter if the sequence name is
    *          unambiguous. If more than one connected database has a sequence
    *          with given name, then you must supply <code>ldbName</code>.
    * @param   newValue
    *          An integer expression assigned as the current value of the
    *          specified sequence. If expression is outside the boundary set
    *          by the initial value (at one end) and the lower limit or upper
    *          limit (at the other end) for the sequence, Progress returns an
    *          error, and the sequence value remains unchanged.
    */
   public static void dynamicSetValue(String seqName,
                                      int64 newValue,
                                      String ldbName)
   {
      getSequenceHandler(ldbName).setValue(seqName,
                                           ldbName,
                                           newValue.longValue());
   }

   /**
    * Maps the <code>DYNAMIC-CURRENT-VALUE(seqName, ldbName) = expr</code> 
    * statement from Progress as it is defined in Open Edge 10.x.
    * Resets the current value of the sequence defined in the Data
    * Dictionary.
    *
    * @param   seqName
    *          An identifier that specifies the name of a sequence defined
    *          in the Data Dictionary.
    * @param   ldbName
    *          An identifier that specifies the logical name of the database
    *          in which the sequence is defined. The database must be
    *          connected. You can omit this parameter if the sequence name is
    *          unambiguous. If more than one connected database has a sequence
    *          with given name, then you must supply <code>ldbName</code>.
    * @param   newValue
    *          An integer expression assigned as the current value of the
    *          specified sequence. If expression is outside the boundary set
    *          by the initial value (at one end) and the lower limit or upper
    *          limit (at the other end) for the sequence, Progress returns an
    *          error, and the sequence value remains unchanged.
    */
   public static void dynamicSetValue(String seqName,
                                      int64 newValue,
                                      character ldbName)
   {
      String _ldbName = ldbName.toJavaType();
      getSequenceHandler(_ldbName).setValue(seqName,
                                            _ldbName,
                                            newValue.longValue());
   }

   /**
    * Maps the <code>DYNAMIC-CURRENT-VALUE(seqName, ldbName) = expr</code>
    * statement from Progress as it is defined in Open Edge 10.x.
    * Resets the current value of the sequence defined in the Data
    * Dictionary.
    *
    * @param   seqName
    *          A character expression that evaluates to the name of a 
    *          sequence defined in the Data Dictionary.
    * @param   ldbName
    *          A character expression that evaluates to the name of a database
    *          in which the sequence is defined. The database must be
    *          connected. You can omit this parameter if the sequence name is
    *          unambiguous. If more than one connected database has a sequence
    *          with given name, then you must supply <code>ldbName</code>.
    * @param   newValue
    *          An integer expression assigned as the current value of the
    *          specified sequence. If expression is outside the boundary set
    *          by the initial value (at one end) and the lower limit or upper
    *          limit (at the other end) for the sequence, Progress returns an
    *          error, and the sequence value remains unchanged.
    */
   public static void dynamicSetValue(character seqName,
                                      long newValue,
                                      String ldbName)
   {
      getSequenceHandler(ldbName).setValue(seqName.toJavaType(), 
                                           ldbName,
                                           newValue);
   }

   /**
    * Maps the <code>DYNAMIC-CURRENT-VALUE(seqName, ldbName) = expr</code> 
    * statement from Progress as it is defined in Open Edge 10.x.
    * Resets the current value of the sequence defined in the Data
    * Dictionary.
    *
    * @param   seqName
    *          An identifier that specifies the name of a sequence defined
    *          in the Data Dictionary.
    * @param   ldbName
    *          An identifier that specifies the logical name of the database
    *          in which the sequence is defined. The database must be
    *          connected. You can omit this parameter if the sequence name is
    *          unambiguous. If more than one connected database has a sequence
    *          with given name, then you must supply <code>ldbName</code>.
    * @param   newValue
    *          An integer expression assigned as the current value of the
    *          specified sequence. If expression is outside the boundary set
    *          by the initial value (at one end) and the lower limit or upper
    *          limit (at the other end) for the sequence, Progress returns an
    *          error, and the sequence value remains unchanged.
    */
   public static void dynamicSetValue(character seqName,
                                      long newValue,
                                      character ldbName)
   {
      String _ldbName = ldbName.toJavaType();
      getSequenceHandler(_ldbName)
         .setValue(seqName.toJavaType(), _ldbName, newValue);
   }

   /**
    * Maps the <code>DYNAMIC-CURRENT-VALUE(seqName, ldbName) = expr</code>
    * statement from Progress as it is defined in Open Edge 10.x.
    * Resets the current value of the sequence defined in the Data
    * Dictionary.
    *
    * @param   seqName
    *          A character expression that evaluates to the name of a 
    *          sequence defined in the Data Dictionary.
    * @param   ldbName
    *          A character expression that evaluates to the name of a database
    *          in which the sequence is defined. The database must be
    *          connected. You can omit this parameter if the sequence name is
    *          unambiguous. If more than one connected database has a sequence
    *          with given name, then you must supply <code>ldbName</code>.
    * @param   newValue
    *          An integer expression assigned as the current value of the
    *          specified sequence. If expression is outside the boundary set
    *          by the initial value (at one end) and the lower limit or upper
    *          limit (at the other end) for the sequence, Progress returns an
    *          error, and the sequence value remains unchanged.
    */
   public static void dynamicSetValue(character seqName,
                                      int64 newValue,
                                      String ldbName)
   {
      getSequenceHandler(ldbName).setValue(seqName.toJavaType(),
                                           ldbName,
                                           newValue.longValue());
   }

   /**
    * Maps the <code>DYNAMIC-CURRENT-VALUE(seqName, ldbName) = expr</code>
    * statement from Progress as it is defined in Open Edge 10.x.
    * Resets the current value of the sequence defined in the Data
    * Dictionary.
    *
    * @param   seqName
    *          A character expression that evaluates to the name of a 
    *          sequence defined in the Data Dictionary.
    * @param   ldbName
    *          A character expression that evaluates to the name of a database
    *          in which the sequence is defined. The database must be
    *          connected. You can omit this parameter if the sequence name is
    *          unambiguous. If more than one connected database has a sequence
    *          with given name, then you must supply <code>ldbName</code>.
    * @param   newValue
    *          An integer expression assigned as the current value of the
    *          specified sequence. If expression is outside the boundary set
    *          by the initial value (at one end) and the lower limit or upper
    *          limit (at the other end) for the sequence, Progress returns an
    *          error, and the sequence value remains unchanged.
    */
   public static void dynamicSetValue(character seqName,
                                      int64 newValue,
                                      character ldbName)
   {
      String _ldbName = ldbName.toJavaType();
      getSequenceHandler(_ldbName)
         .setValue(seqName.toJavaType(), _ldbName, newValue.longValue());
   }

   /**
    * Seek the database in which the sequence is defined. It is assumed that 
    * there should no conflicts or the first found database is returned.
    * This method is called when some (static) sequence function has been
    * called without the last parameter (ldbName).
    * As a side effect, if the database is found, it is guaranteed that the
    * appropriate <code>Sequence</code> handler has been loaded.
    *
    * @param   seqName
    *          The name of the sequence to lookup after.
    *
    * @return The name of the database where the sequence is defined.
    */
   private static String getDatabaseForSequence(String seqName)
   {
      int connectedDbCount = ConnectionManager.numDBs().toJavaType();
      for (int i = 0; i < connectedDbCount; i++)
      {
         String ldbName = ConnectionManager.ldbName(i + 1).toJavaType();
         SequenceHandler handler = getSequenceHandler(ldbName);
         if (handler != null && handler.contains(ldbName, seqName))
         {
            return ldbName;
         }
      }

      return null;
   }

   /**
    * Factory method for obtaining a sequence object for a database.  If some
    * settings is not found in directory a <code>SequenceHandler</code> is
    * returned by default.
    *
    * @param   ldbName
    *          The name of the database
    *
    * @return  An instance of Sequence that will to which the internal
    *          functionality will be delegated.
    */
   private static SequenceHandler getSequenceHandler(String ldbName)
   {
      synchronized (SequenceManager.class)
      {
         if (seqHandlers.containsKey(ldbName))
         {
            // sequence handler already prepared for this database
            return seqHandlers.get(ldbName);
         }

         SequenceHandler sequenceHandler = null;
         Persistence persistence = ConnectionManager.getPersistence(ldbName);
         Dialect dialect = persistence.getDialect();
         
         /* Detect the appropriate sequence handler based on the Dialect
            associated with the database */
         if (dialect instanceof P2JH2Dialect)
         {
            sequenceHandler = new H2SequenceHandler();
         }
         else if (dialect instanceof P2JPostgreSQLDialect)
         {
            sequenceHandler = new PostgreSQLSequenceHandler();
         }
         else 
         {
            // Is there a new database or an unknown dialect ?
            sequenceHandler = new SequenceHandler();
            System.err.println("Could not detect dialect for " + ldbName);
         }

         // save to internal cache for further fast access
         seqHandlers.put(ldbName, sequenceHandler);
         return sequenceHandler;
      }
   }

   /**
    * Creates the internal data structure representing a sequence object and 
    * add it to in-memory list. 
    * This method is only called at startup time by <code>DMOIndex</code> when
    * it is initializing (from <code>updateMap()</code>).
    * 
    * @param   schema
    *          The schema/database.
    * @param   seqName
    *          The name of the sequence.
    * @param   seqInc
    *          The increment value of the sequence.
    * @param   seqMin
    *          The minimum value of the sequence.
    * @param   seqMax
    *          The maximum value of the sequence.
    * @param   seqCycle
    *          The cycle flag of the sequence.
    */
   public static void addSequenceDefinition(String  schema,
                                            String  seqName,
                                            long    seqInc,
                                            long    seqMin,
                                            long    seqMax,
                                            boolean seqCycle)
   {
      // first check if Sequence already exists
      Sequence sequence = getSequence(schema, seqName);
      if (sequence != null)
      {
         return;
      }

      // sequence does not exists, create it now
      sequence = new Sequence(schema,
                              seqName,
                              seqMin,
                              seqMax,
                              seqInc,
                              seqCycle);

      HashMap<String, Sequence> dbSeq = null;
      if (sequences.containsKey(schema))
      {
         // if database exists, use it
         dbSeq = sequences.get(schema);
      }
      else
      {
         // otherwise, create it now
         dbSeq = new HashMap<String, Sequence>();
         sequences.put(schema, dbSeq);
      }

      // add the object to database list
      dbSeq.put(seqName, sequence);
   }

   /**
    * Lookups the internal data structure representing a sequence object.
    * 
    * @param   ldbName
    *          The logical name of the database.
    * @param   name
    *          The name of the sequence.
    */
   public static Sequence getSequence(String ldbName, String name)
   {
      if (!sequences.containsKey(ldbName))
      {
         return null;
      }

      HashMap<String, Sequence> dbSeqList = sequences.get(ldbName);
      if (dbSeqList == null || !dbSeqList.containsKey(name))
      {
         return null;
      }

      return dbSeqList.get(name);
   }
}
